""" add_srcdirs_to_syspath module. """
import re
import sys
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import Generator, List, Optional, Pattern, Set, Tuple, Union

_STD_SYSPATH_FILTER: Union[None, Pattern] = None


def init_std_syspath_filter(std_syspath_filter: Pattern) -> None:
    """
    Provide a globally bound standard filter Pattern applied to all subsequent sys.path filtering
    operations.  Note: a init_std_syspath_filter() is called from __init__.py to default to a
    globally applied pattern.
    :param std_syspath_filter: pattern to apply to all filter operations. Can be None.
    :return: None
    """
    # pylint: disable=global-statement
    global _STD_SYSPATH_FILTER
    # pylint: enable=global-statement
    _STD_SYSPATH_FILTER = std_syspath_filter


def filtered_sorted_syspath(
    path_filter: Pattern = None, no_filtering: bool = False, sort: bool = False
) -> List[str]:
    """
    Filter and sort the syspath for only paths of interest.
    :param path_filter: a pattern that the user can provide in addition to the std_syspath_filter
    :param no_filtering: allow user to not filter at all
    :param sort: allow user to sort the filtered (if filtering) sys.path
    :return: sys.path with filtering and sorting applied
    """
    paths: List[str] = sys.path
    if not no_filtering:
        path: str
        if _STD_SYSPATH_FILTER:
            paths = [path for path in paths if not re.search(_STD_SYSPATH_FILTER, path)]
        if path_filter:
            paths = [path for path in paths if not re.search(path_filter, path)]

    return sorted(paths, reverse=True) if sort else paths


def print_syspath(
    path_filter: Pattern = None, no_filtering: bool = False, sort: bool = True
) -> None:
    """
    Filter and sort the syspath for only paths of interest.
    :param path_filter: a pattern that the user can provide in addition to the std_syspath_filter
    :param no_filtering: allow user to not filter at all
    :param sort: allow user to sort the filtered (if filtering) sys.path
    :return: None
    """
    paths: List[str] = filtered_sorted_syspath(path_filter, no_filtering, sort)
    print(f"\nsys.path({len(paths)} paths):")
    path: str
    for path in paths:
        print(f"\t{path}")


def add_srcdirs_to_syspath() -> None:
    """
    Add all src directories under current working directory to sys.path. If CWD is not the
    project root (containing root 'src' directory), walk up ancestry to find a directory
    containing a 'src' directory. Waking up allows for being within in the 'tests' directory
    when initiating tests against modules under 'root/src'.

    Searching for 'src' directories is NOT limited to finding the '<project root>/src' (and 'src'
    directories under that '<project root>/src' directory)! All those will be found and added,
    but also any other 'src' directory found under the <project root>/tests tree. This is desired
    since git subprojects may be under 'tests' and their 'src' directories need to be
    included.

    :return: None
    """
    root_path: Path = Path.cwd()

    while not (root_path / "src").exists() or not (root_path / "src").is_dir():
        if not root_path.parent:
            raise RuntimeError(
                f"The CWD {Path.cwd().as_posix()} (and any parents of that path) do not contain a "
                f"'src' directory."
            )
        root_path = root_path.parent

    if root_path != Path.cwd():
        print(f"Searching for 'src' dirs from {root_path.as_posix()}")

    prior_sys_path = sys.path.copy()

    all_src_dirs: Generator[Path, None, None] = root_path.glob("src")
    all_test_src_dirs: Generator[Path, None, None] = root_path.glob("tests/**/src")

    tested_src: Path
    for tested_src in chain.from_iterable([all_src_dirs, all_test_src_dirs]):
        tested_src_str = str(tested_src)
        if tested_src.is_dir() and tested_src_str not in sys.path:
            sys.path.append(tested_src_str)

    diff_path_strs: Set[str] = set(prior_sys_path).symmetric_difference(set(sys.path))
    if len(diff_path_strs) > 0:
        diff_path_strs = {Path(diff_path_str).as_posix() for diff_path_str in diff_path_strs}
        print(f"Added to sys.path: {sorted(diff_path_strs)}")


def get_package_and_max_relative_import_dots(
    module_name: str,
) -> Tuple[Optional[str], Optional[str]]:
    """
    Derive the fully-qualified package related to already-imported module named by 'module_name'.
    In addition, return the number of relative dots that can be used in that module before either of
    the following occur:

    ValueError: attempted relative import beyond top-level package
    ImportError: attempted relative import beyond top-level package

    :param module_name: module name of already-imported module
    :return: fully-qualified package and max relative dots.
    """
    target_module: ModuleType = sys.modules[module_name]
    dots: str = "" if not target_module.__package__ else "."
    dot_count: int = target_module.__package__.count(".") if target_module.__package__ else 0
    dots: str = dots + "".join("." for i in range(0, dot_count))
    return target_module.__package__, dots



""" __init__ module. """
import re

from .syspath_utils import (
    add_srcdirs_to_syspath,
    filtered_sorted_syspath,
    get_package_and_max_relative_import_dots,
    init_std_syspath_filter,
    print_syspath,
)

__version__ = "0.1.35"

init_std_syspath_filter(re.compile(r"([Jj]et[Bb]rains|[Pp]ython|PyCharm|v\w*env)"))



import importlib
import inspect
import logging
import shutil
import site
import sys
from importlib import reload
from pathlib import Path
from typing import List, Optional, Tuple

import click
import importlib_metadata
from diff_match_patch import diff_match_patch, patch_obj

from . import syspath_sleuth
from .syspath_sleuth import SysPathSleuth

PRE_SLEUTH_SUFFIX = ".pre_sleuth"
REVERSE_PATCH_SUFFIX = ".patch"

error_logger: logging.Logger = logging.getLogger(f"{__name__}.error")
error_logger.addHandler(logging.StreamHandler(sys.stderr))

sleuth_logger: logging.Logger = logging.getLogger(error_logger.parent.name)
sleuth_logger.addHandler(logging.StreamHandler(sys.stdout))


class InstallError(RuntimeError):
    pass


class UninstallError(RuntimeError):
    pass


def append_sleuth_to_customize(customize_path: Path, syspath_sleuth_path: Optional[Path] = None):
    sleuth_logger.info(
        "Appending %s to site customize: %s",
        *get_name_and_relative_path(customize_path, syspath_sleuth_path),
    )

    lines: List[str]

    if syspath_sleuth_path:
        with syspath_sleuth_path.open() as custom_syspath_sleuth:
            lines = custom_syspath_sleuth.readlines()
    else:
        lines, _ = inspect.getsourcelines(syspath_sleuth)

    with customize_path.open("r+") as customize_path_f:
        customize_path_f.writelines(lines)


def create_site_customize(customize_path: Path):
    sleuth_logger.info("Creating system site %s", customize_path.name)
    customize_path.touch()


def copy_site_customize(customize_path: Path):
    shutil.copy(customize_path, customize_path.with_suffix(PRE_SLEUTH_SUFFIX))


def create_reverse_sleuth_patch(customize_path):
    pre_sleuth_customize_path = customize_path.with_suffix(PRE_SLEUTH_SUFFIX)
    if customize_path.exists and pre_sleuth_customize_path.exists:
        customize: str
        with customize_path.open() as customize_f:
            customize = customize_f.read()

        pre_sleuth_site_customize: str
        with pre_sleuth_customize_path.open() as pre_sleuth_customize_f:
            pre_sleuth_site_customize = pre_sleuth_customize_f.read()

        dmp = diff_match_patch()
        diffs: List[patch_obj] = dmp.diff_main(
            customize, pre_sleuth_site_customize, checklines=False
        )
        patches = dmp.patch_make(diffs)
        customize_reverse_patch: Path = customize_path.with_suffix(REVERSE_PATCH_SUFFIX)
        reverse_patch: str = dmp.patch_toText(patches)

        with customize_reverse_patch.open("x") as customize_reverse_patch_f:
            customize_reverse_patch_f.writelines(reverse_patch)

        pre_sleuth_customize_path.unlink()


def reverse_patch_sleuth(customize_path):
    reverse_patch_path = customize_path.with_suffix(REVERSE_PATCH_SUFFIX)
    if not reverse_patch_path.exists():
        return

    sleuth_logger.info(
        "Removing %s from site customize: %s",
        SysPathSleuth.__name__,
        SysPathSleuth.relative_path(customize_path),
    )
    with reverse_patch_path.open() as customize_patch_f:
        patch = customize_patch_f.read()

        dmp = diff_match_patch()
        patches: List[str] = dmp.patch_fromText(patch)

    patched_customize: str
    patch_results: List[bool]
    with customize_path.open("r") as customize_patch_f:
        customize = customize_patch_f.read()
        patched_customize, patch_results = dmp.patch_apply(patches, customize)
        save_patched = bool(patched_customize)
        for patch_result in patch_results:
            if not patch_result:
                raise UninstallError(
                    f"Reverse patch failed; patch file: "
                    f"{reverse_patch_path}.\n"
                    f"Hand edit removal of {SysPathSleuth.__name__}"
                )
        if save_patched:
            with customize_path.open("w") as customize_patch_f:
                customize_patch_f.seek(0)
                customize_patch_f.write(patched_customize)

    reverse_patch_path.unlink()
    if not save_patched:
        customize_path.unlink()

    try:
        # pylint: disable=import-outside-toplevel
        import sitecustomize

        # pylint: enable=import-outside-toplevel

        sys.path = sys.path.get_base_list()
        if isinstance(sys.path, sitecustomize.SysPathSleuth):
            error_logger.warning("Hmmm... expected sys.path NOT to be monkey-patched.")

    except (AttributeError, ModuleNotFoundError):
        # This will occur if SysPathSleuth was not installed prior. But, don't skip the
        # uninstall_sleuth() as the user messaging associated with this condition is shared.
        pass


def get_user_customize_path():
    return Path(site.getusersitepackages()) / "usercustomize.py"


def get_system_customize_path() -> Path:
    system_site: str
    for system_site in site.getsitepackages():
        if "site-packages" in system_site:
            return Path(system_site) / "sitecustomize.py"
    raise InstallError("No system site found!")


def get_name_and_relative_path(
    customize_path: Path, syspath_sleuth_path: Optional[Path]
) -> Tuple[str, Path]:
    if syspath_sleuth_path:
        sleuth_name = syspath_sleuth_path.name
        sleuth_path: Path = get_relative_path(syspath_sleuth_path)
    else:
        sleuth_name = SysPathSleuth.__name__
        sleuth_path: Path = SysPathSleuth.relative_path(customize_path)
    return sleuth_name, sleuth_path


def get_relative_path(path: Path) -> Path:
    sleuth_path: Path = path
    try:
        sleuth_path = path.relative_to(Path.cwd())
    except ValueError:
        pass
    return sleuth_path


def inject_sleuth(syspath_sleuth_path: Optional[Path] = None):

    # When using venv, site.ENABLE_USER_SITE is False. When using virtual environments,
    # the effort is to isolate the activities within one virtual environment per Python
    # system Python from other virtual environments. Were the user site enabled, it would
    # affect other Python virtual environments.
    user_path = False
    if site.ENABLE_USER_SITE and site.check_enableusersite():
        customize_path = get_user_customize_path()
        customize_path.parent.mkdir(parents=True, exist_ok=True)
        user_path = True
    else:
        customize_path = get_system_customize_path()

    if customize_path and customize_path.exists():
        name, _ = get_name_and_relative_path(customize_path, syspath_sleuth_path)
        sleuth_logger.warning(
            "Reinstalling %s in %s site...", name, "user" if user_path else "system"
        )
        reverse_patch_sleuth(customize_path)

    create_site_customize(customize_path)
    copy_site_customize(customize_path)
    append_sleuth_to_customize(customize_path, syspath_sleuth_path)
    create_reverse_sleuth_patch(customize_path)

    # Determine if the customize site was updated to wrap sys.path with a SysPathSleuth.
    if site.ENABLE_USER_SITE and site.check_enableusersite():
        customize_module = importlib.import_module("usercustomize")
    else:
        customize_module = importlib.import_module("sitecustomize")
    reload(customize_module)
    class_names: Tuple[str] = tuple(
        x[0] for x in inspect.getmembers(customize_module, inspect.isclass)
    )
    if syspath_sleuth_path and (
        "SysPathSleuth" not in class_names
        or not isinstance(sys.path, customize_module.SysPathSleuth)
    ):
        # The file loaded doesn't wrap sys.path with a SysPathSleuth
        sleuth_logger.setLevel(logging.ERROR)
        reverse_patch_sleuth(customize_path)
        relative_path = get_relative_path(syspath_sleuth_path)
        raise InstallError(f"{relative_path} does not wrap sys.path with a SysPathSleuth.")


def uninstall_sleuth():
    # When using venv, site.ENABLE_USER_SITE is False. When using virtual environments,
    # the effort is to isolate the activities within one virtual environment per Python
    # system Python from other virtual environments. Were the user site enabled, it would
    # affect other Python virtual environments.
    user_path = False
    if site.ENABLE_USER_SITE and site.check_enableusersite():
        user_path = True
        customize_path = get_user_customize_path()
    else:
        customize_path = get_system_customize_path()

    if not customize_path.exists():
        error_logger.warning(
            "%s was not installed in %s site: %s",
            SysPathSleuth.__name__,
            "user" if user_path else "system",
            SysPathSleuth.relative_path(customize_path),
        )
        return

    reverse_patch_sleuth(customize_path)

    sleuth_logger.warning(
        "%s uninstalled from %s site: %s",
        SysPathSleuth.__name__,
        "user" if user_path else "system",
        SysPathSleuth.relative_path(customize_path),
    )


@click.command(
    help="(Un)Install SysPathSleuth into user-site or system-site to track sys.path "
    "access in real-time."
)
@click.version_option(version=importlib_metadata.version("runtime-syspath"))
@click.option("--inject/--uninstall", "-i/-u", default=False, help="default=uninstall")
@click.option(
    "--custom",
    "-c",
    type=click.Path(exists=True, resolve_path=True),
    help="path to a user's implementation of a SysPathSleuth",
)
@click.option("--verbose", "-v", is_flag=True, default=False)
def syspath_sleuth_main(inject: bool, custom: str, verbose: bool):
    custom_path: Optional[Path] = Path(custom) if custom else None
    if verbose:
        sleuth_logger.setLevel(logging.INFO)
        for handler in sleuth_logger.handlers:
            handler.setLevel(logging.INFO)

    try:
        if inject:
            inject_sleuth(custom_path)

            # handler = logging.StreamHandler(sys.stdout)
            # handler.setLevel(logging.INFO)
            # sys.path.config_logger(handler, logging.INFO)
            # sys.path.append('yow')
        else:
            try:
                # pylint: disable=import-outside-toplevel
                import sitecustomize

                # pylint: enable=import-outside-toplevel

                sys.path = sys.path.get_base_list()
                if isinstance(sys.path, sitecustomize.SysPathSleuth):
                    error_logger.warning("Hmmm... expected sys.path NOT to be monkey-patched.")

            except (AttributeError, ModuleNotFoundError):
                # This will occur if SysPathSleuth was not installed prior. But, don't skip the
                # uninstall_sleuth() as the user messaging associated with this condition is shared.
                pass

            uninstall_sleuth()
    except Exception as e:
        error_logger.error(f"{'Inject' if inject else 'Uninstall'} failed: {e}")



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['runtime_syspath', 'runtime_syspath.syspath_sleuth']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'diff-match-patch>=20200713,<20200714',
 'importlib-metadata>=2.0.0,<3.0.0']

entry_points = \
{'console_scripts': ['syspath_sleuth_injector = '
                     'runtime_syspath.syspath_sleuth.__main__:syspath_sleuth_main']}

setup_kwargs = {
    'name': 'runtime-syspath',
    'version': '0.2.7',
    'description': "Functions show and find each 'src' directory under working directory and add each to sys.path.",
    'long_description': '`runtime-syspath` is a package to ease programmatically adding src root\npaths to `sys.path`. This is targeted at python test code that needs to\ndiscover a project\'s solution source to test.\n\n> :exclamation: It is generally **frowned upon** to alter the `sys.path`\n> programmatically as it confuses development, especially refactoring.\n> Python IDEs can statically determine if a dependent package\'s import\n> statement is left wanting whether a PyPi installation in needed or\n> source cannot be discovered through standard Python paths. A static\n> analysis tool\'s *missing import* detection will end up registering\n> false-negatives if the import is discovered via dynamic (programmatic)\n> additions to `sys.path` at runtime.\n\n*The following description assumes the use of `pytest` unit testing\nsupport and a project file structuring that includes project root\ndirectories named `src` (project solution) and `tests` (project tests of\nproject source under `src`. Both `src` and `tests` are not intended to\nhave package initializers (`__init__.py`). Packages therein will\ntypically have package initializers allowing for test modules to have\nthat same name (in separate packages). However, as a general rule, test\nmodules are not intended to import other test modules. Therefore, there\nshould be no need for `__init__.py`-enabled, relative importation\nbetween test cases or sub-package test cases. `pytest`\'s\n[default test discovery](https://docs.pytest.org/en/latest/goodpractices.html#test-discovery)\nand intended design use negates the need for :*\n\n```\n├─ src\n│  └─ __init__.py\n|  └─ foo.py\n├─ tests\n│  └─ test_foo.py\n│     └─ foo_and_goo\n│        └─ __init__.py\n│        └─ test_foo.py\n│        └─ test_goo.py\n└─ setup.py\n```\n*That structure is based upon\n[this guidance](https://blog.ionelmc.ro/2014/05/25/python-packaging/#the-structure).*\n\nWhen testing solution source in a project, the test cases _could_\nstatically access the solution source by importing with the `src`\npackage prefix:\n\n```\nimport src.packagename.foo\n```\nNot only does that not feel right at all, that solution implies that\ntests are run **only** from the project root, not within the `tests`\ndirectory itself. If the test is run within the `tests` directory, the\n`src` package won\'t be found at runtime.\n\nSo, using:\n```\nimport packagename.foo\n```\n... the `src` directory would need to be programmatically added to the\n`sys.path`. This will allow for tests to be run form any working\ndirectory under the `tests` sub-tree.\n\n`runtime_syspath.add_srcdirs_to_syspath()` will discover all `src`\ndirectories under `<project root>/src`. The reason that there may be\nmore is if your project may be leveraging `git subprojects` under\n`<project root>/src` that have their own `src` directories. Those need\nto be added to `sys.path` also.\n\nTo leverage `runtime-syspath` to add the `src` directory everytime a\ntest is run, import `runtime-syspath` and run\n`add_srcdirs_to_syspath()` in `tests/conftest.py`. (If `tests`\ncontain more `conftest.py` under its directory tree, the call still only\nneed appear in the root `test/conftest.py`!):\n ```\n from runtime_syspath import add_srcdirs_to_syspath\n \n add_srcdirs_to_syspath() \n ```\n\n`add_srcdirs_to_syspath()` will recursively discover **all** `src`\nsubdirectories under the <project root>. For projects that use `git\nsubmodules`, their `src` directories need to be added to `src.path` for\nimport access. `git subprojects` could be added to `src` or `tests`\ndirectory trees:\n\n```\n├─ src\n│  └─ __init__.py\n|  └─ projectpackage\n│     └─ __init__.py\n|     └─ foo.py\n|  └─ subproject\n|     └─ src\n│       └─ __init__.py\n|       └─ bar.py\n|     └─ tests\n├─ tests\n│  └─ test_foo.py\n|  └─ test_subproject\n|     └─ src\n│       └─ __init__.py\n|       └─ unfoobarrator.py\n|     └─ tests\n└─ setup.py\n```\n\n> :exclamation: Due to the code maintenance and grok\'ing mayhem caused\n> by indiscriminate runtime additions to `sys.path`, your goal should be\n> to limit that anti-pattern to this discovery-of-source aspect for\n> import discovery.\n\n> :bulb: Since programmatically adding to a `sys.path` impairs an IDE\'s\n> ability to do static import discovery and leveraging IDE refactoring\n> features between the solution source and the test code, an IDE user\n> would need to manually mark all `src` directories as such.  \n> PyCharm example:\n>\n> ![image](docs/images/IDE_SetSrc.png)\n\n#### SysPathSleuth; runtime reporting of programmatic `sys.path` access\n\nOn a project riddled with programmatically appending source paths to\n`sys.path`, a tool to discover which modules are mucking with `sys.path`\nand when could prove useful. This discovery can assist with manually\neradicating `sys.path` access in favor of updating imports with\nfully-qualified (anchored at but, not including `src`), absolute\nmodule/package names. static tools would then be able to discover the\nmodules/packages imported.\n> Relative paths: There is a place for relative paths when importing\n> intra-package modules. But, when importing inter-package modules,\n> leveraging fully-qualified, absolute module/package names is a wiser\n> play.\n\nSysPathSleuth is a monkey-patch of `sys.path` to report on `sys.path`\naccess that comes with an installer to install/uninstall SysPathSleuth\ninto either the user or system site\'s _customize_ modules\n(`~/pathto/user_site/usercustomize.py` or\n`/pathto/python/site-packages/sitecustomize.py`). SysPathSleuth can be\ninstalled/uninstalled using:\n* python -m syspath_slueth \\[--install _or_ --uninstall]\n* at the start within a running program\n\nAt the start of a running program prior:\n```\nimport atexit\nimport syspath_sleuth\nfrom runtime-syspath import syspath_slueth\n\nsyspath_sleuth.inject_sleuth()\ndef uninstall_syspath_sleuth():\n    syspath_sleuth.uninstall_sleuth()\n\natexit.register(uninstall_syspath_sleuth)\n\nif __name__ == "__main__":\n    go_main_go()\n\n```\n',
    'author': 'Greg Kedge',
    'author_email': 'gregwork@kedges.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gkedge/runtime-syspath',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import inspect
import logging
import site
import sys
from pathlib import Path, PurePath
from types import FrameType
from typing import List, Optional


class SysPathSleuth(list):
    logger: logging.Logger = logging.getLogger("runtime-syspath.SysPathSleuth")
    logger.setLevel(logging.NOTSET)
    logger.propagate = False

    def insert(self, *args):
        self._where("insert", args)
        return super().insert(*args)

    def append(self, *args):
        self._where("append", args)
        return super().append(*args)

    def extend(self, *args):
        self._where("extend", args)
        return super().extend(*args)

    def pop(self, *args):
        self._where("pop", args)
        return super().pop(*args)

    def remove(self, *args):
        self._where("remove", args)
        return super().remove(*args)

    def __delitem__(self, *args):
        self._where("__delitem__", args)
        return super().__delitem__(*args)

    def __setitem__(self, *args):
        self._where("__setitem__", args)
        return super().__setitem__(*args)

    def get_base_list(self) -> List[str]:
        return list(self)

    @classmethod
    def config_logger(cls, handler: logging.Handler = None, level: int = -1):
        if level != -1:
            cls.logger.setLevel(level)
        if handler:
            cls.logger.addHandler(handler)
            if cls.logger.getEffectiveLevel() < handler.level:
                logger_level_name = logging.getLevelName(cls.logger.getEffectiveLevel())
                handler_level_name = logging.getLevelName(handler.level)
                message = (
                    f"Handler's level ({handler_level_name}) is insufficient "
                    f"to log at {cls.logger.name}'s level ({logger_level_name}) yet."
                )
                cls._inform_user(message)

    @classmethod
    def _where(cls, action, args):

        frame_info: Optional[inspect.Traceback] = None
        # Only inspect the slooow stack introspection if print()'ing or logging level is sufficient.
        if not cls._is_logging_on() or cls.logger.isEnabledFor(logging.INFO):
            syspath_caller: FrameType = inspect.currentframe().f_back.f_back
            if not (inspect.istraceback(syspath_caller) or inspect.isframe(syspath_caller)):
                return
            frame_info: inspect.Traceback = inspect.getframeinfo(syspath_caller)

        if frame_info:
            filename = PurePath(frame_info.filename)
            try:
                filename = PurePath(frame_info.filename).relative_to(sys.base_prefix)
            except ValueError:
                filename_orig = filename
                cwd = Path.cwd()
                while filename == filename_orig:
                    try:
                        filename = PurePath(frame_info.filename).relative_to(cwd)
                        break
                    except ValueError:
                        cwd = cwd.parent

            message = f"sys.path.{action}{args} from {filename}:{frame_info.lineno}"
            cls._inform_user(message)

    @classmethod
    def _is_logging_on(cls):
        is_logging_on = cls.logger.getEffectiveLevel() != logging.NOTSET
        if is_logging_on:
            handler: logging.Handler
            for handler in cls.logger.handlers:
                if handler.level != logging.NOTSET:
                    break
            else:
                is_logging_on = False
        return is_logging_on

    @classmethod
    def _inform_user(cls, message: str, logging_level: int = logging.INFO):
        if cls._is_logging_on() and cls.logger.isEnabledFor(logging_level):
            cls.logger.log(logging_level, message)
        else:
            print(message)

    @classmethod
    def is_sleuth_active(cls):
        sleuth_module_file_name = PurePath(__file__).name
        # Do both so report informs user whether it has been installed both.
        is_active_in_user_site = cls._is_active_in_user_site(sleuth_module_file_name)
        is_active_in_system_site = cls._is_active_in_system_site(sleuth_module_file_name)

        return is_active_in_user_site or is_active_in_system_site

    @staticmethod
    def get_user_customize_path() -> Path:
        user_customize_module_name = PurePath("usercustomize.py")
        user_customize_path = Path(site.getusersitepackages()) / user_customize_module_name
        return user_customize_path

    @staticmethod
    def get_system_customize_path() -> Optional[Path]:
        site_customize_module_name = PurePath("sitecustomize.py")

        system_site: str
        for system_site in site.getsitepackages():
            system_site_path: Path = Path(system_site)
            if system_site_path.name == "site-packages" and system_site_path.is_dir():
                return system_site_path / site_customize_module_name
        return None

    @staticmethod
    def relative_path(customize_path):
        try:
            customize_path = customize_path.relative_to(sys.base_prefix)
        except ValueError:
            customize_path_orig = customize_path
            cwd = Path.cwd()
            while customize_path == customize_path_orig:
                try:
                    customize_path = customize_path.relative_to(cwd)
                    break
                except ValueError:
                    if str(cwd) == "/":
                        break
                    cwd = cwd.parent
        return customize_path

    @classmethod
    def _is_active_in_user_site(cls, sleuth_module_file_name):
        # When using venv, site.ENABLE_USER_SITE is False. When using virtual environments,
        # the effort is to isolate the activities within one virtual environment per Python
        # system Python from other virtual environments. Were the user site enabled, it would
        # affect other Python virtual environments.
        if not site.ENABLE_USER_SITE or not site.check_enableusersite():
            return False

        if sleuth_module_file_name != "usercustomize.py":
            # It is likely syspath_sleuth.py is being tested
            return False

        user_customize_path = cls.get_user_customize_path()
        if not user_customize_path.exists():
            return False
        user_customize_path = cls.relative_path(user_customize_path)
        sleuth_message = f"SysPathSleuth is installed in user site: {user_customize_path}"
        cls._inform_user(sleuth_message)
        return True

    @classmethod
    def _is_active_in_system_site(cls, sleuth_module_file_name):
        if sleuth_module_file_name != "sitecustomize.py":
            # It is likely syspath_sleuth.py is being tested
            return False

        system_custom_path = cls.get_system_customize_path()
        if not system_custom_path or not system_custom_path.exists():
            return False

        system_custom_path = cls.relative_path(system_custom_path)
        sleuth_message = f"SysPathSleuth is installed in system site: {system_custom_path}"
        cls._inform_user(sleuth_message)
        return True


# Might be pytest'ing...
if SysPathSleuth.is_sleuth_active():
    sys.path = SysPathSleuth(sys.path)
